/*
Operservice 2 - trusts.c
Manages Trusts for hosts or groups of hosts
(C) Michael Meier 2000-2001 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "globals.h"

array deniedtrusts; /* List of denied trusts */
int defaultclonewarn=6;  // Default for warning about clones (6 on quakenet, 0 to disable)
trustedhost * trustedhosts[SIZEOFTL];       // The trusted hosts
array untrustedhosts[SIZEOFHL];     // normal, untrusted hosts
trustedgroup * trustedgroups[SIZEOFIDMAP];
int glineonclones=0;

unsigned int hlhash(char *host) {
  return (unsigned int)(crc32(host)%SIZEOFHL);
}

unsigned int tlhash(char *host) {
  return (unsigned int)(crc32(host)%SIZEOFTL);
}

unsigned int tgshash(unsigned long ID) {
  return (unsigned int)(ID%SIZEOFIDMAP);
}

/* Returns 0 if host isn't trusted, or the ID of the trusted host group */
unsigned long istrusted(char *host) {
  trustedhost *a;
  a=trustedhosts[tlhash(host)];
  while (a!=NULL) {
    if (strcmp(host,a->hostname)==0) { return a->id; }
    a=(void *)a->next;
  }
  return 0;
}

int addtountrustedhost(char *host) {
  unsigned int n; unsigned long i; hostcounter *a;
  n=hlhash(host);
  a=(hostcounter *)(untrustedhosts[n].content);
  for (i=0;i<untrustedhosts[n].cursi;i++) {
    if (strcmp(host,a[i].hostname)==0) {
      a[i].currenton++;
      return a[i].currenton;
    }
  }
  i=array_getfreeslot(&(untrustedhosts[n]));
  a=(hostcounter *)(untrustedhosts[n].content);
  a[i].hostname=getastring(host);
  a[i].currenton=1;
  return 1;
}

void delfromuntrustedhost(char *host) {
  unsigned int n; unsigned long i; hostcounter *a;
  n=hlhash(host);
  a=(hostcounter *)(untrustedhosts[n].content);
  for (i=0;i<untrustedhosts[n].cursi;i++) {
    if (strcmp(host,a[i].hostname)==0) {
      a[i].currenton--;
      if (a[i].currenton==0) {
        freeastring(a[i].hostname);
        array_delslot(&(untrustedhosts[n]),i);
      }
      return;
    }
  }
}

void addtotrustedhost(char *ident, unsigned long id) {
  trustedgroup *tg; char tmps2[TMPSSIZE]; unsigned long i; identcounter *ic;
  tg=findtrustgroupbyID(id);
  if (tg==NULL) { return; }
  tg->currentlyon++;
  if (tg->currentlyon > tg->trustedfor) {
    sprintf(tmps2,"Trustgroup %s exceeding trustlimit - %d / %d",tg->name,tg->currentlyon,tg->trustedfor);
    if (burstcomplete!=0) { noticeallircops(tmps2); }
    putlog("[trustexceeded] %s",tmps2);
  }
  if (strlen(ident)>USERLEN) { return; }
  ic=(identcounter *)(tg->identcounts.content);
  for (i=0;i<tg->identcounts.cursi;i++) {
    if (strcmp(ic[i].ident,ident)==0) {
      ic[i].currenton++;
      if ((ic[i].currenton > tg->maxperident) && (tg->maxperident!=0)) {
        sprintf(tmps2,"User %s @ Group %s has excessive connections (%d / %d)",ident,tg->name,ic[i].currenton,tg->maxperident);
        putlog("[excon] %s",tmps2);
        if (burstcomplete!=0) { noticeallircops(tmps2); }
      }
      return;
    }
  }
  i=array_getfreeslot(&(tg->identcounts));
  ic=(identcounter *)(tg->identcounts.content);
  ic[i].currenton=1;
  strcpy(ic[i].ident,ident);
}

void delfromtrustedhost(char *ident, unsigned long id) {
  trustedgroup *tg; unsigned long i; identcounter *ic;
  tg=findtrustgroupbyID(id);
  if (tg==NULL) { return; }
  tg->currentlyon--;
  if (strlen(ident)>USERLEN) { return; }
  ic=(identcounter *)(tg->identcounts.content);
  for (i=0;i<tg->identcounts.cursi;i++) {
    if (strcmp(ic[i].ident,ident)==0) {
      ic[i].currenton--;
      if (ic[i].currenton==0) {
        array_delslot(&(tg->identcounts),i);
      }
      return;
    }
  }
}

void trustnewclient(char *ident, char *host) {
  char tmph[HOSTLEN+1]; unsigned long thid; char tmps2[TMPSSIZE];
  if (strlen(ident)>USERLEN) { return; }
  if (strlen(host)>HOSTLEN) { return; }
  strcpy(tmph,host); toLowerCase(tmph);
  /* First check if this host is trusted */
  thid=istrusted(tmph);
  if (thid!=0) { /* That host is trusted */
    addtotrustedhost(ident, thid);
  } else { /* untrusted host */
    int currenton;
    currenton=addtountrustedhost(tmph);
    if (currenton>=defaultclonewarn) { /* Clonelimit for that host exceeded */
      if (defaultclonewarn==0) { return; }
      if (burstcomplete!=0) {
        if (glineonclones==0) {
          sprintf(tmps2,"[%d] clones detected from %s",currenton,host);
        } else {
          sprintf(tmps2,"[%d] clones detected from %s - HOST GLINED!",currenton,host);
        }
        noticeallircops(tmps2);
        if (glineonclones==1) {
          sprintf(tmps2,"*@%s",host);
          addgline(tmps2,"Too many clones from your host","CloneCheck",30*60,1);
        }
      }
      putlog("[clones] %d clones detected from untrusted host %s",currenton,host);
    }
  }
}

void trustdelclient(char *ident, char *host) {
  char tmph[HOSTLEN+1]; unsigned long thid;
  strcpy(tmph,host); toLowerCase(tmph);
  /* First check if this host is trusted */
  thid=istrusted(tmph);
  if (thid!=0) { /* That host is trusted */
    delfromtrustedhost(ident, thid);
  } else { /* untrusted host */
    delfromuntrustedhost(tmph);
  }
}

trustedgroup * findtrustgroupbyID(unsigned long ID) {
  trustedgroup *tmp;
  tmp=trustedgroups[tgshash(ID)];
  while (tmp!=NULL) {
    if (ID==tmp->id) {
      return tmp;
    }
    tmp=(void *)tmp->next;
  }
  return NULL;
}

trustedgroup * findtrustgroupbyname(char *name) {
  trustedgroup *tmp; int i;
  for (i=0;i<SIZEOFIDMAP;i++) {
    tmp=trustedgroups[i];
    while (tmp!=NULL) {
      if (strcmp(name,tmp->name)==0) {
        return tmp;
      }
      tmp=(void *)tmp->next;
    }
  }
  return NULL;
}

unsigned long getfreetgid() {
  unsigned long id; trustedgroup *tg;
  id=0;
  do {
    id++;
    tg=findtrustgroupbyID(id);
  } while (tg!=NULL);
  return id;
}

int destroytrustgroup(unsigned long id) {
  array listofhosts; trustedhost *th; long i, j; trustedhost *th2, *th3;
  trustedgroup *tg1, *tg2, *tg3; userdata *u; char tmph[HOSTLEN+1]; char tmps2[TMPSSIZE];
  /* Not only will we have to delete the trustgroup, we will also have to recount
     all individual hosts in it. So lets start with building a list of all hosts */
  array_init(&listofhosts,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    th2=NULL;
    while (th!=NULL) {
      if (th->id==id) {
        /* This host belongs to our trustgroup */
        long tmpl=array_getfreeslot(&listofhosts);
        strcpy(listofhosts.content+tmpl*(HOSTLEN+1),th->hostname);
        if (th2==NULL) {
          trustedhosts[i]=(void *)th->next;
        } else {
          th2->next=th->next;
        }
        th3=th;
        th=(void *)th->next;
        free(th3);
      } else {
        th2=th;
        th=(void *)th->next;
      }
    }
  }
  /* we now have the list in array listofhosts, and the hosts deleted from trustedhosts */
  /* Goodbye trustgroup */
  tg1=trustedgroups[tgshash(id)]; tg2=NULL;
  while (tg1!=NULL) {
    if (tg1->id==id) { /* Kill it! */
      array_free(&(tg1->identcounts));
      if (tg2==NULL) {
        trustedgroups[tgshash(id)]=(void *)tg1->next;
      } else {
        tg2->next=tg1->next;
      }
      tg3=tg1;
      tg1=(void *)tg1->next;
      free(tg3);
    } else {
      tg2=tg1;
      tg1=(void *)tg1->next;
    }
  }
  /* Recount all hosts - this is gonna be a CPU Killer, but it's used less than
     once a day, so i don't care too much. */
  for (i=0;i<SIZEOFUL;i++) {
    u=uls[i];
    while (u!=NULL) {
      strcpy(tmph,u->host); toLowerCase(tmph);
      for (j=0;j<listofhosts.cursi;j++) {
        if (strcmp(tmph,listofhosts.content+j*(HOSTLEN+1))==0) {
          int currenton=addtountrustedhost(tmph);
          if (currenton>=defaultclonewarn) { /* Clonelimit for that host exceeded */
            if (defaultclonewarn!=0) {
              sprintf(tmps2,"[%d] clones detected from (previously trusted host) %s",currenton,u->host);
              noticeallircops(tmps2);
              putlog("[clones] %d clones detected from (now) untrusted host %s",currenton,u->host);
            }
          }
        }
      }
      u=(void *)u->next;
    }
  }
  i=listofhosts.cursi;
  array_free(&listofhosts);
  return i;
}

/* Trusts have to be loaded BEFORE linking to the net.
   this function returns the number of hosts loaded from the database. */
unsigned long loadtrusts() {
  int res; MYSQL_RES *myres; MYSQL_ROW myrow; unsigned long hostcounter=0;
  res=mysql_query(&sqlcon,"SELECT * FROM trustgroups");
  if (res!=0) {
    putlog("!!! Failed to query trustgroups from database !!!");
    return 0;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=9) {
    putlog("!!! trustgroups in database have invalid format !!!");
    return 0;
  }
  while ((myrow=mysql_fetch_row(myres))) {
    if ((strlen(myrow[1])>TRUSTNAMELEN) || (strlen(myrow[6])>TRUSTCONTACTLEN) ||
        (strlen(myrow[7])>TRUSTCOMLEN) || (strlen(myrow[8])>AUTHUSERLEN)) {
      putlog("!!! Malformed trustgroups-entry in database! !!!");
    } else {
      if (strtol(myrow[0],NULL,10)==0) {
        putlog("!!! Trustgroups-ID of 0 in database! !!!");
      } else {
        trustedgroup *tg; unsigned int thehash;
        tg=malloc(sizeof(trustedgroup));
        tg->id=strtol(myrow[0],NULL,10);
        strcpy(tg->name,myrow[1]);
        tg->trustedfor=strtol(myrow[2],NULL,10);
        tg->expires=strtol(myrow[3],NULL,10);
        tg->maxperident=strtol(myrow[4],NULL,10);
        tg->enforceident=strtol(myrow[5],NULL,10);
        strcpy(tg->contact,myrow[6]);
        strcpy(tg->comment,myrow[7]);
        strcpy(tg->creator,myrow[8]);
        array_init(&(tg->identcounts),sizeof(identcounter));
        tg->currentlyon=0;
        thehash=tgshash(tg->id);
        tg->next=(void *)trustedgroups[thehash];
        trustedgroups[thehash]=tg;
        hostcounter++;
      }
    }
  }
  mysql_free_result(myres);
  hostcounter=hostcounter<<16;
  res=mysql_query(&sqlcon,"SELECT * FROM trustedhosts");
  if (res!=0) {
    putlog("!!! Failed to query trustedhosts from database !!!");
    return hostcounter;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=2) {
    putlog("!!! trustedhosts in database have invalid format !!!");
    return hostcounter;
  }
  while ((myrow=mysql_fetch_row(myres))) {
    if (strlen(myrow[0])>HOSTLEN) {
      putlog("!!! Malformed trustedhosts-entry in database! !!!");
    } else {
      trustedgroup *tg; unsigned int thehash; trustedhost *th;
      tg=findtrustgroupbyID(strtol(myrow[1],NULL,10));
      if (tg==NULL) {
        putlog("!!! trustedhosts: Referential integrity violated! (no group #%ld)",strtol(myrow[1],NULL,10));
      } else {
        th=malloc(sizeof(trustedhost));
        strcpy(th->hostname,myrow[0]);
        th->id=strtol(myrow[1],NULL,10);
        thehash=tlhash(myrow[0]);
        th->next=(void *)trustedhosts[thehash];
        trustedhosts[thehash]=th;
        hostcounter++;
      }
    }
  }
  mysql_free_result(myres);
  trustgroupexpire();
  return hostcounter;
}

void savetrusts() {
  char sqlquery[5000]; long i; trustedgroup *tg; trustedhost *th;
  char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  /* No need to write expired data, so expire old shit first */
  trustgroupexpire();
  /* Clear the tables */
  mysql_query(&sqlcon,"DELETE FROM trustgroups");
  mysql_query(&sqlcon,"DELETE FROM trustedhosts");
  for (i=0;i<SIZEOFIDMAP;i++) {
    tg=trustedgroups[i];
    while (tg!=NULL) {
      mysql_escape_string(tmps2,tg->name,strlen(tg->name));
      mysql_escape_string(tmps3,tg->contact,strlen(tg->contact));
      mysql_escape_string(tmps4,tg->comment,strlen(tg->comment));
      mysql_escape_string(tmps5,tg->creator,strlen(tg->creator));
      sprintf(sqlquery,"INSERT INTO trustgroups VALUES(%lu, '%s', %d, %lu, %d, %d, '%s', '%s', '%s')",
        tg->id,tmps2,tg->trustedfor,tg->expires,tg->maxperident,tg->enforceident,tmps3,tmps4,tmps5);
      mysql_query(&sqlcon,sqlquery);
      tg=(void *)tg->next;
    }
  }
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      mysql_escape_string(tmps2,th->hostname,strlen(th->hostname));
      sprintf(sqlquery,"INSERT INTO trustedhosts VALUES('%s', %lu)",tmps2,th->id);
      mysql_query(&sqlcon,sqlquery);
      th=(void *)th->next;
    }
  }
}

void trustgroupexpire() {
  long i; trustedgroup *tg; unsigned long j;
  for (i=0;i<SIZEOFIDMAP;i++) {
    tg=trustedgroups[i];
    while (tg!=NULL) {
      if (tg->expires<getnettime()) {
        j=tg->id;
        tg=(void *)tg->next;
        destroytrustgroup(j);
      } else {
        tg=(void *)tg->next;
      }
    }
  }
}

trustdeny * gettrustdeny(char *host) {
  trustdeny *td; long i; char tmps2[TMPSSIZE];
  td=(trustdeny *)deniedtrusts.content;
  mystrncpy(tmps2,host,TMPSSIZE-1);
  for (i=0;i<deniedtrusts.cursi;i++) {
    if (match2strings(td[i].host,tmps2)) {
      return &td[i];
    }
  }
  return NULL;    
}

/* this function returns the number of deniedtrusts loaded from the database. */
unsigned long loadtrustdeny() {
  int res; MYSQL_RES *myres; MYSQL_ROW myrow;
  res=mysql_query(&sqlcon,"SELECT * FROM trustdenys");
  if (res!=0) {
    putlog("!!! Failed to query trustdenys from database !!!");
    return 0;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=6) {
    putlog("!!! trustdenys in database have invalid format !!!");
    return 0;
  }
  while ((myrow=mysql_fetch_row(myres))) {
    if ((strlen(myrow[0])>HOSTLEN) || (strlen(myrow[1])>AUTHUSERLEN) ||
        (strlen(myrow[2])>RNGREAS)) {
      putlog("!!! Malformed trustdeny-entry in database! !!!");
    } else {
      /* FIX ME */
      long i; trustdeny * td;
      i=array_getfreeslot(&deniedtrusts);
      td=(trustdeny *)deniedtrusts.content;
      mystrncpy(td[i].host,myrow[0],HOSTLEN);
      mystrncpy(td[i].creator,myrow[1],AUTHUSERLEN);
      mystrncpy(td[i].reason,myrow[2],RNGREAS);
      td[i].expires=strtoul(myrow[3],NULL,10);
      td[i].created=strtoul(myrow[4],NULL,10);
      if (strtol(myrow[5],NULL,10)==1) {
        td[i].type=TRUSTDENY_DENY;
      } else {
        td[i].type=TRUSTDENY_WARN;
      }
    }
  }
  mysql_free_result(myres);
  trustdenyexpire();
  return 0;
}

void savetrustdeny() {
  char sqlquery[4*TMPSSIZE+500]; long i; trustdeny *td;
  char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  /* No need to write expired data, so expire old shit first */
  trustdenyexpire();
  /* Clear the tables */
  mysql_query(&sqlcon,"DELETE FROM trustdenys");
  td=(trustdeny *)deniedtrusts.content;
  for (i=0;i<deniedtrusts.cursi;i++) {
    mysql_escape_string(tmps2,td[i].host,strlen(td[i].host));
    mysql_escape_string(tmps3,td[i].creator,strlen(td[i].creator));
    mysql_escape_string(tmps4,td[i].reason,strlen(td[i].reason));
    sprintf(sqlquery,"INSERT INTO trustdenys VALUES('%s','%s','%s','%lu','%lu','%d')",
      tmps2,tmps3,tmps4,td[i].expires,td[i].created,td[i].type);
    mysql_query(&sqlcon,sqlquery);
  }
}

void trustdenyexpire() {
  long i; trustdeny * td;
  for (i=0;i<deniedtrusts.cursi;i++) {
    td=(trustdeny *)deniedtrusts.content;
    if ((td[i].expires!=0) && (td[i].expires<getnettime())) {
      array_delslot(&deniedtrusts,i);
      i--;
    }
  }
}
